#/usr/bin/env python
#!/usr/bin/env python
#coding:utf-8
# Author:   Yaron Butterfield
# Purpose: For RNASeq-Align.  If this cigar (after alignment to genome + junctions) 
#          contains S, I, or D tags, convert the cigar to include junction regions
#    
# Created: 7/20/2011 / original 2009/2010

import sys
import re   

def read_cigar(cigar):
        splitlist = re.split('([MIDNSHP])', cigar)
        newlist = [splitlist[x]+splitlist[x+1] for x in range(0, len(splitlist)-1, 2)]
        return newlist


def getCigarValues(cigar):

        c = read_cigar(cigar)
        c1 = []
        for k in c:
                c1.append([int(k[:-1]),k[-1]])        
       
        totalCov=0
        for cov in c1:
                if cov[1] =='M':
                        totalCov += cov[0]
                elif cov[1] == 'N':
                        totalCov += cov[0]
                elif cov[1] == 'D':
                        totalCov += cov[0]
                elif cov[1] == 'I':
                        pass
                elif cov[1] == 'S':
                        pass

        return totalCov

def cigarCheck(nC1,c,seqLength):

        Mcount=0 # number of M, S, or I aligned bases; must equal sequence length

        #print "--------------------------------------------||||||||||||||||||||---->",c

        nC = read_cigar(nC1)

        if "-" in nC1:
		return 0

        if nC[0][-1]=='S' and nC[1][-1]=='N':
                return 0

        for k in c:
                if k[-1]=='M':
                        Mcount += int(k[:-1])
                elif k[-1]=='S':
                        Mcount += int(k[:-1])
                elif k[-1]=='I':
                        Mcount += int(k[:-1])

        if Mcount <> seqLength:
                #print "?",Mcount, seqLength
                return 0
        else:
                return 1


def run_100mer_tests():
        seqlength=100


        origCigars = ["5M1D95M","41M1I58M", "63M1I36M", "9M1I90M",
                      "61M1I38M", "57M1D43M", "81M1I18M", "94M2D4M1I1M",
                      "67M2I31M", "1M1I98M", "14M5I81M", "93M2D1M1I5M", 
                      "67M2I31M", "1M2D1M1I97M", "92M2I6M", "6M1D1M2I91M",
                      "92M3I3M2D2M", "94M3I2M2D1M", "95M1D4M1D1M", "3S4M2D93M",
                      "91M3I5M2D1M", "3M3I3M2D91M", "95M3D3M2S", "9S91M", 
	              "92M8S", "92M3D4M4S"]
        tempCigars = ["7M528N83M545N10M","22M24066N50M1803N28M", "44M24066N50M1803N6M", "40M1806N60M",
                      "42M24066N50M1806N8M", "38M24066N50M1803N12M", "36M2630N62M220N2M", "76M1839N18M6S",
                      "1S98M2732N1M", "7S56M165N37M", "14M9450N84M7725N2M", "69M89N31M", 
                      "1S98M2732N1M", "96M1612N4M", "80M9340N13M7S", "8S5M1822N87M",
                      "91M9065N9M", "14M1506N80M4409N6M", "72M236N28M", "8S32M2423N60M",
                      "7M16364N85M38867N8M", "8M106N92M", "21M691N78M1S", "2M20N2M20N2M20N44M20N50M",
	              "48M40N45M20N2M20N2M20N3M", "1M123N96M103N2M1S"]
        correctCigars = ["5M1D1M528N83M545N11M","22M24066N19M1I31M1803N27M", "44M24066N19M1I31M1803N5M", "9M1I31M1806N59M",
                         "42M24066N19M1I31M1806N7M", "38M24066N19M1D30M1803N13M", "36M2630N45M1I17M220N1M", "76M1839N18M6S",
                         "1S66M2I31M", "8S56M165N36M", "14M9450N5I81M", "69M89N24M2D1M1I5M", 
                         "1S66M2I31M", "1M2D1M1I92M1612N5M", "80M9340N12M1I7S", "8S1I5M1822N86M",
                         "91M9065N1M3I3M2D2M", "14M1506N80M4409N3I2M2D1M", "72M236N23M1D4M1D1M", "7S1D31M2423N62M",
                         "7M16364N84M3I1M38867N4M2D1M", "3M3I2M106N1M2D91M", "21M691N74M3D3M2S", "9S41M20N50M",
	                 "48M40N44M8S", "1M123N91M3D4M4S"]

        c = 0

        if len(origCigars) <> len(tempCigars) and len(tempCigars) <> len(correctCigars) and len(correctCigars) <> len(origCigars):
                print "Number of CIGARS are not matching"
                exit(0)

        totalCorrect = 0
        incorrect=[]
        for i in origCigars:

                #print "Example:"

                zcigar = origCigars[c]
                cigar = tempCigars[c]
                correct = correctCigars[c]

                (newcigar, a, b, intr) = convert_new(cigar,zcigar)
                testNumber = "Test " + str(c+1)
                print "                                 ",testNumber
                print "z.cigar",zcigar
                print "cigar", cigar
                print "new",newcigar
                print "cor",correct
		print "intron bool", intr
		
                if correct==newcigar:
                        totalCorrect+=1
                        print " :D  Yes!"
                else:
                        print " :(  Oops"
                        incorrect.append(testNumber)

                print '----------------------------------------'

                c+=1


        print "Total examined:",str(len(origCigars))
        print "Total correct cigars:",str(totalCorrect)
        if len(origCigars) == totalCorrect:
                print "All cigars are correct!!"
        else:
                print "Incorrect cigar(s):", incorrect


def run_75mer_tests():
        seqlength=75
        # 75 mers 


        origCigars = ["26M4I45M","4M3I68M","50M1D25M","26M4I45M","8M1I66M",
                      "3M1D72M","5M1D70M","5M1I69M","46M3D29M","6M1D69M",
                      "16M3I56M","68M1I6M","71M1D4M","4S71M","45M3D30M","18M2D57M","5M1D70M",
                      "4S1M3D70M"]

        tempCigars = ["26M3078N49M","8M1500N25M2340N42M",'4S71M','26M3078N49M',"7M815N68M",
                      "30M1000N45M","6S60M150N9M","6S60M150N9M","3M1677N44M2019N28M","7S68M",
                      "19M2175N56M","67M8S","18M24240N52M5S","73M837N2M","47M3487N28M","1S73M857N1M","7M528N68M",
                      "7S68M"]

        correctCigars = ["26M3078N4I45M","4M3I4M1500N25M2340N39M","4S46M1D25M","26M3078N4I45M","7M815N1M1I66M",
                         "3M1D26M1000N46M","6S60M150N9M","6S60M150N9M","3M1677N44M2019N3D28M","7S68M",
                         "16M3I3M2175N53M","67M8S","18M24240N57M","4S69M837N2M","45M3D3487N30M","1S17M2D56M857N1M","5M1D1M528N69M",
                         "7S1D68M"]

        c=0

        if len(origCigars) <> len(tempCigars) and len(tempCigars) <> len(correctCigars) and len(correctCigars) <> len(origCigars):
                print "Number of CIGARS are not matching"
                exit(0)

        totalCorrect = 0
        incorrect=[]
        for i in origCigars:

                #print "Example:"

                zcigar = origCigars[c]
                cigar = tempCigars[c]
                correct = correctCigars[c]

                (newcigar, a, b, intr) = convert_new(cigar,zcigar)
                testNumber = "Test " + str(c+1)
                print "                                 ",testNumber
                print "z.cigar",zcigar
                print "cigar", cigar
                print "new",newcigar
                print "cor",correct

                if correct==newcigar:
                        totalCorrect+=1
                        print " :D  Yes!"
                else:
                        print " :(  Oops"
                        incorrect.append(testNumber)

                print '----------------------------------------'

                c+=1


        print "Total examined:",str(len(origCigars))
        print "Total correct cigars:",str(totalCorrect)
        if len(origCigars) == totalCorrect:
                print "All cigars are correct!!"
        else:
                print "Incorrect cigar(s):", incorrect



def get_ZJ_tag(tempCigar, pos, intron_bool):
        # custom feature requested by Nina Theissen which lists first and last bases of intron(s) skipped
        # Uses information from repositioned coordinate, final cigar, and temp cigar
	
	#ZJ is calculated using temp cigar (repositionied cigar before merging with S's D's and I's), and 
	#is then adjusted to remove introns that are in softclips using the information stored in intron_bool
        """
         14786,'44M140N6M1000N25M'
        """
	
	#initialize
        t = read_cigar(tempCigar)  # this is the temp cigar where the introns were inserted, before merging with S's D's and I's
    #    f = read_cigar(finalCigar) 

        ZJtag=''
        exonAligned=0
        intronSize=0
        pairs=[]
	
	#iterate through elements of temp cigar, calculating coordinates of each intron 
	#since zj tag is being calculated using the temp cigar, we're using the initial pos calculated by "reposition" (the 'start' variable)
	#this pos is updated visavis softclips from the jaguar alignment, hence 'S' elements are not included in the iteration
	#the final update of the pos due to bwa softclips happens after the zj tag is calculated
	
	for element in t:

                if element[-1]=='N':
                        intronSize = int(element[:-1])

                        pairs.append(str(pos + exonAligned)+"-"+str(pos + exonAligned + intronSize -1))
                        pos = pos + exonAligned + intronSize 

                        exonAligned =0
               
                elif element[-1]=='M':
                        exonAligned += int(element[:-1])
					
	
	#remove any introns that were removed from final cigar due to softclipping. 
	if 0 in intron_bool:
		pairs = [pairs[i] for i in range(len(intron_bool)) if intron_bool[i]==1]
        
	#compile ZJ tag	
        ZJtag = 'ZJ:Z:'+ ','.join(pairs)
	if not pairs:  #this happens if the only junction in the read was removed due to softclipping or insertion
		ZJtag = 'ZJ:Z:G'
		
        return ZJtag	


def run_ZJ_tests():

        tests = [[14786, '44M140N6M', '44M140N6M', [1]],
                 [14786, '44M140N6M1000N25M', '44M140N6M1000N25M', [1,1]],
                 [44047613, '8M528N83M545N9M', '5M1D2M528N83M545N10M', [0,1]],
	         [200, '1S98M2732N1M', '1S66M2I31M', [0]],
	         [203, '1S52M200N33M300N14M', '3S50M200N33M300N14M', [1,1]]
                 ]

        for i in tests:
                pos = i[0]
                tempcigar = i[1]
		finalcigar = i[2]
		int_bool = i[3]

                ZJtag = get_ZJ_tag(tempcigar, finalcigar, pos, int_bool)	
                print ZJtag


def convert_new(cigar, zcigar):

        C = read_cigar(cigar)
        J = read_cigar(zcigar)


        #######################
        #process original cigar, J

        #get distance, from beginning, of D's and I's
        cumulative = 0
        DeeEyes = [] #collects locations of D's and I's from start of read
        for interval in J:
                if not re.search("[DI]", interval):
                        interval_len = int(re.match("(\d+)", interval).group(1))
                        cumulative = cumulative + interval_len
                else:
                        res = re.match("(\d+)([DI])$", interval)
                        interval_len = int(res.group(1))
                        interval_type = res.group(2)
                        DeeEyes.append([interval_type, cumulative+1, interval_len]) #array containing D/I, start of event, event length
                        if interval_type == 'I':
                                cumulative = cumulative + interval_len


        #get S's in original cigar

        Ess_beg = ''
        Ess_end = ''

        if 'S' in J[0]:
                Ess_beg  = int(re.match("(\d+)", J[0]).group(1))    
        if 'S' in J[-1]:
                Ess_end = int(re.match("(\d+)", J[-1]).group(1))


        # add or merge S's from original cigar to temp cigar array, C. If large gaps are removed as a result, record this in intron boolean
	intron_bool = [1]* len(re.findall("N", cigar)) # initialize intron boolean
	add_to_pos = 0  #records number of N bases removed only because of start softclips
	aa=0
	
	if Ess_beg and ('S' in C[0]) and (Ess_beg > int(re.search("(\d+)", C[0]).group(1))):  #if there are Ss in both original and temp cigars, and the S from the original cigar is longer than the one from the temp cigar
        #NOTE: if there are Ss in both original and temp cigars and S from the original cigar is SHORTER than the one from the temp cigar then C does not need to be updated        
                S_length_tempcigar = int(re.search("(\d+)", C[0]).group(1))
		M_length_tempcigar = int(re.search("(\d+)", C[1]).group(1))
		extra_clipped_bases = 0
		num_removed = 0
                while M_length_tempcigar <= (Ess_beg - (S_length_tempcigar+extra_clipped_bases)): # if the S from the original cigar is longer than the one from the temp cigar, add the longest to C
                        #drop the preceding M and N, so we propagate the clipping up past the N to the next-to-last M element in the tempcigar			
			num_removed +=1
			intron_bool[num_removed -1] = 0
			del C[1] # remove next M
			add_to_pos += int(re.search("(\d+)", C[1]).group(1)) #record length of N being removed
			del C[1] #remove next N
			extra_clipped_bases += M_length_tempcigar
			M_length_tempcigar = int(re.search("(\d+)", C[1]).group(1)) # get ready to test M in the next while loop                        
                add_to_pos -= S_length_tempcigar
                #once we get this far, we can safely shorten the first M       
		C[0] = str(Ess_beg) + 'S'
		M_length_tempcigar -= Ess_beg - (S_length_tempcigar+extra_clipped_bases)
		C[1] = str(M_length_tempcigar) + 'M' 	#adjust length of next M	
		aa=1

        elif Ess_beg and ('S' not in C[0]): #if there is an S in the original cigar but none in the temp cigar, add the origS to C	
                Sstring = J[0]		
                C.insert(0, Sstring) #add an S element to first position of C
                M_length_tempcigar = int(re.search("(\d+)", C[1]).group(1))
                extra_clipped_bases = 0		
		num_removed = 0
                while M_length_tempcigar <= Ess_beg - extra_clipped_bases:  #if softclip is longer than next M
		        num_removed +=1
			intron_bool[num_removed -1] = 0		        
                        del C[1] # remove next M
			add_to_pos += int(re.search("(\d+)", C[1]).group(1)) #record length of N being removed
                        del C[1] #remove next N
			extra_clipped_bases += M_length_tempcigar
                        M_length_tempcigar = int(re.search("(\d+)", C[1]).group(1))  
		M_length_tempcigar -= Ess_beg - extra_clipped_bases #adjust length of next M	
                C[1] = str(M_length_tempcigar) + 'M'  
		aa=1
		
        if Ess_end and ('S' in C[-1]) and (Ess_end > int(re.search("(\d+)", C[-1]).group(1))): #if there are Ss in both original and temp cigars, and the S from the original cigar is longer than the one from the temp cigar
                S_length_tempcigar = int(re.search("(\d+)", C[-1]).group(1))
                M_length_tempcigar = int(re.search("(\d+)", C[-2]).group(1))
                extra_clipped_bases = 0
		num_removed = 0
                while M_length_tempcigar <= (Ess_end - (S_length_tempcigar+extra_clipped_bases)): #the last M in tempcigar is not big enough to absorb all the extra softclipping we need to account for
                        #drop the preceding M and N, so we propagate the clipping up past the N to the next-to-last M element in the tempcigar
			num_removed +=1
			intron_bool[-num_removed] = 0			
                        del C[-2] # remove previous M
                        del C[-2] #remove previous N
                        extra_clipped_bases += M_length_tempcigar
                        M_length_tempcigar = int(re.search("(\d+)", C[-2]).group(1)) # get ready to test M in the next while loop
                #once we get this far, we can safely shorten the last M
                C[-1] = str(Ess_end) + 'S' 
                M_length_tempcigar -= Ess_end - (S_length_tempcigar+extra_clipped_bases)	
                C[-2] = str(M_length_tempcigar) + 'M' #adjust length of previous M	

        elif Ess_end and ('S' not in C[-1]): #if there is an S in the original cigar but none in the temp cigar, add the S to C	
                Sstring = str(Ess_end) + 'S'		
                C.append(Sstring) #add an S element to last position of C
                M_length_tempcigar = int(re.search("(\d+)", C[-2]).group(1))
                extra_clipped_bases = 0  
		num_removed = 0
                while M_length_tempcigar <= Ess_end - extra_clipped_bases: #if softclip is longer than previous M
			num_removed +=1
			intron_bool[-num_removed] = 0	                        
                        del C[-2] # remove previous M
                        del C[-2] #remove previous N
			extra_clipped_bases += M_length_tempcigar
                        M_length_tempcigar = int(re.search("(\d+)", C[-2]).group(1)) 
		M_length_tempcigar -= Ess_end - extra_clipped_bases #adjust length of pevious M		
                C[-2] = str(M_length_tempcigar) + 'M'

        #######################
        #iterate over each piece of the temporary, gapped cigar, claculating locations of I/D events in each exon

        exons2end_of_curr_M = 0
        exons2start_of_current_M = 0
        to_add = 0 
        cuml_string =[]
        dropped_indels_num = 0
        dropped_indels_bases = 0

        n=0  #counts number of intervals
        for interval in C:
                n = n+1
		try:
			interval_len = int(re.match("(\d+)", interval).group(1))
		except:
			print "ABORTING. Error processing interval='%s' in C='%s' (full cigar='%s', J='%s', zcigar='%s')" % (interval, C, cigar, J, zcigar)
			sys.exit(-1)
			
                if "S" in interval:
                        exons2end_of_curr_M += interval_len
                        S_size = interval_len
                        if n==1: # deal with softclip in beginning of cigar	
                                while DeeEyes: #I/D events, we are always in the [0] position because of the pop 
                                        event_end = DeeEyes[0][1] + DeeEyes[0][2] -1
                                        if event_end <= interval_len: #if an entire I/D event is in the softclip region, remove it
                                                dropped_indels_num += 1
                                                dropped_indels_bases += DeeEyes[0][2]
                                                if DeeEyes[0][0]=='I':
                                                        to_add -= DeeEyes[0][2]
                                                        S_size += DeeEyes[0][2]
                                                elif DeeEyes[0][0]=='D':
                                                        to_add += DeeEyes[0][2]
                                                        S_size -= DeeEyes[0][2]
                                                DeeEyes.pop(0)
                                        elif DeeEyes[0][1] <= interval_len and event_end > interval_len: #if part of an I/D event is in a softclip region, remove that part and keep the remaining part
                                                new_event_size = event_end - interval_len
                                                orig_event_size = DeeEyes[0][2]
                                                dropped_indels_num += 1
                                                event_size_diff = orig_event_size - new_event_size
                                                dropped_indels_bases += event_size_diff 
                                                if DeeEyes[0][0]=='I':
                                                        to_add -= event_size_diff
                                                        S_size += event_size_diff
                                                elif DeeEyes[0][0]=='D':  #this is not working properly, but very rare. reads go into cigar.txt and are unmapped.
                                                        to_add += event_size_diff
                                                        S_size -= event_size_diff
                                                new_event_start = S_size + 1
                                                DeeEyes[0][1] = new_event_start
                                                DeeEyes[0][2] = new_event_size                                                
                                                break
                                        else: # if there's no D/I event in the softclip region, break D/I event loop
                                                break        
                                cuml_string.append(str(S_size) + 'S')   
                        else: #deal with softclip at end of cigar                                            
                                while DeeEyes: #I/D events, we are always in the [0] position because of the pop 
                                        if DeeEyes[0][1] <= exons2end_of_curr_M: 
                                                dropped_indels_num += 1
                                                dropped_indels_bases += DeeEyes[0][2] 
                                                DeeEyes.pop(0)
                                cuml_string.append(str(S_size) + 'S')   #if there's an ID event beyond the beginning of the softclip, it is ignored
                        
                elif "M" in interval:
			if aa==1:
				interval_len+=Ess_beg
				to_add -= Ess_beg
				aa=0
				
                        lastflag = 0
                        cuml_MsbeforeEvent = 0
                        to_change = 0
                        I_len = 0
                        exons2start_of_current_M = exons2end_of_curr_M
                        exons2end_of_curr_M += interval_len

                        while DeeEyes: #loop over I/D events in this exon; we are always in the [0] position because of the pop 
                                if DeeEyes[0][1] <= exons2end_of_curr_M: 

                                        #calculate location of event with respect to start of current exon, or with respect to last event in the same exon     
                                        Mlength_before = DeeEyes[0][1] - exons2start_of_current_M - cuml_MsbeforeEvent -I_len -1
                                        cuml_MsbeforeEvent += Mlength_before

                                        #make different accounting for D's and I's
                                        if DeeEyes[0][0]=='D':		    
                                                to_add += DeeEyes[0][2] #for D these need to be added to last M,
                                                to_change += DeeEyes[0][2]
                                        else:
                                                to_add -= DeeEyes[0][2] #for I these need to be subtracted from the last M
                                                I_len += DeeEyes[0][2]
                                        M_string_before = str(Mlength_before) + 'M'
                                        if M_string_before == '0M':
                                                M_string_before = ''
                                        event_len = DeeEyes[0][2]
                                        event_type = DeeEyes[0][0]
                                        event_string  = str(event_len) + event_type
                                        cuml_string.append(M_string_before + event_string) #string for the entire exon so far
                                        DeeEyes.pop(0) #so that we don't get events that have been covered in previous M's

                                        if DeeEyes: # if there is a next D/I event
                                                if DeeEyes[0][1] <= exons2end_of_curr_M: #look ahead to see if next D/I event is in the same interval
                                                        continue
                                                else: 
                                                        lastflag = 1 #this is not the last event, but it is the last event in this interval
                                        else:
                                                lastflag = 1 #this is the last event in this read

                                        if lastflag == 1:
                                                M_last = interval_len - cuml_MsbeforeEvent - to_change
                                                if n == len(C) or ((n == len(C)-1) and ('S' in C[-1])):  # if this is the last exon, add/subtract extra bases to make up for indels
                                                        M_last += to_add                                                   
                                                if M_last == 0: #if the D/I event end is exactly flush with exon end, no need for Mlast
                                                        if  event_type == 'I' or event_type=='D':
                                                                break
                                                if M_last < 0: #This should only occur for Ds: if event extends beyond this exon, add all of it to the end of this exon, but make correction to accounting. don't write a second M
                                                        if event_type=='D':
								event_remainder = abs(M_last)	
								to_add -= event_remainder	
                                                        break #no need for Mlast here     
                                                M_string_last = str(M_last) + 'M'
                                                cuml_string.append(M_string_last)
                                                break  #exit loop on D/I events once you are in the last one within the current interval

                                else:  #if there are additional DI events, but in a later exon	    
                                        cuml_string.append(str(interval_len) + 'M')
                                        break
                        else:  # this is an else of the while loop: enters if there are no more D/I events in this read
                                if n == len(C) or ((n == len(C)-1) and ('S' in C[-1])): #in case this is the final M
                                        interval_len += to_add	
					num_removed = 0
                                        while interval_len < 0:          #cases of negative final M's   
                                                old_Mlen  = int(re.match("(\d+)", cuml_string[-2]).group(1))  #adjust length on previous M
                                                interval_len  = old_Mlen + interval_len    
                                                cuml_string[-2:] = []  #get rid of last N and previous M
						num_removed+=1
						intron_bool[-num_removed] = 0	
                                        if interval_len==0:
                                                cuml_string[-1:] = []
						intron_bool[-1] = 0
                                                break  #no need to add a 0M
	
                                cuml_string.append(str(interval_len) + 'M')		    

                elif "N" in interval:
                        cuml_string.append(interval)    

        newCigar = "".join(cuml_string)    
        return newCigar, dropped_indels_num, dropped_indels_bases, intron_bool, add_to_pos


if __name__ == '__main__': 
        #run_75mer_tests()
        #run_100mer_tests()
        run_ZJ_tests()
