#!/usr/bin/env python
#coding:utf-8
# Author:   Yaron Butterfield
# Purpose:  Describe SAM tools flag code
# Created: 11/23/2009

import sys
from optparse import OptionParser

"""
 Flag Chr Description 
 
 0x0001 p the read is paired in sequencing 
 0x0002 P the read is mapped in a proper pair 
 0x0004 u the query sequence itself is unmapped 
 0x0008 U the mate is unmapped 
 0x0010 r strand of the query (1 for reverse) 
 0x0020 R strand of the mate 
 0x0040 1 the read is the first read in a pair 
 0x0080 2 the read is the second read in a pair 
 
 0x0100 s the alignment is not primary 
 0x0200 f the read fails platform/vendor quality checks 
 0x0400 d the read is either a PCR or an optical duplicate 

"""

def flags():
    flagDesc={1:"the read is paired in sequencing",
           2:"the read is mapped in a proper pair",
           4:"the query sequence itself is unmapped",
           8:"the mate is unmapped",
           16:"strand of the query (1 for reverse) / read reverse strand",
           32:"strand of the mate / mate reverse strand", # mate is on reverse strand
           64:"the read is the first read in a pair",
           128:"the read is the second read in a pair",
           256:"not primary alignment",
           512:"read fails platform/vendor quality checks", # such as chastity filtering
           1024:"read is PCR or optical duplicate"
           }
    
    flagType={65:[1,64],
              67:[1,2,64],
            69:[1,4,64], 
            73:[1,8,64],
            77:[1,4,8,64], # both reads unaligned
            81:[1,16,64],
            83:[1,2,16,64],
            85:[1,4,16,64],
            87:[1,2,4,16,64],
            89:[1,8,16,64],
            97:[1,32,64],
            99:[1,2,32,64],
            101:[1,4,32,64],
            113:[1,16,32,64], 
            115:[1,2,16,32,64],
            117:[1,4,16,32,64], 
            129:[1,128],
            131:[1,2,128],
            133:[1,4,128],
            137:[1,8,128],
            141:[1,4,8,128], # both reads unaligned
            145:[1,16,128],
            147:[1,2,16,128],
            149:[1,4,16,128],
            151:[1,2,4,16,128],
            153:[1,8,16,128], 
            161:[1,32,128],
            163:[1,2,32,128],
            165:[1,4,32,128],
            177:[1,16,32,128], 
            179:[1,2,16,32,128],
            181:[1,4,16,32,128],
            }
    
    # the following is required to pickup flags used by original pre-repositioned SAM but which
    # we have decided is not valid
    invalidFlagType={93:[1,4,8,16,64],
                     157:[1,4,8,16,128],
                     103:[1,2,4,32,64], #
                     167:[1,2,4,32,128] #
                     }
    flagType.update(invalidFlagType)
    
    moreCode={}
    for flag in flagType:
        
        flagCode = flagType[flag]
        
        newFlag = flag + 512
        newFlagCode = flagCode + [512]
        moreCode[newFlag] = newFlagCode
        
        newFlag = flag + 1024
        newFlagCode = flagCode + [1024]
        moreCode[newFlag] = newFlagCode
        
        newFlag = flag + 512 + 1024
        newFlagCode = flagCode + [512,1024]
        moreCode[newFlag] = newFlagCode
        
        """        
        newFlag=flag + 512
        flagCode = flagType[flag]
        flagCode.append(512)        
        moreCode[newFlag]=flagCode
        
        newFlag=flag + 1024
        flagCode = flagType[flag]
        flagCode.append(1024)        
        moreCode[newFlag]=flagCode

        newFlag=flag + 512+1024
        flagCode = flagType[flag]
        flagCode.append(512)
        flagCode.append(1024)
        moreCode[newFlag]=flagCode
        """
    flagType.update(moreCode)
    
    return flagDesc,flagType
    
       
       
def flagDesc():
    flagDesc,flagType = flags()
    
    for i in flagType:
        print i,
        desc = flagType[i]
        flag = desc(tot)
        print flag

        
def flagInfo(flag):
    flagDesc,flagType = flags()
    flagInfo = []
    desc=''
    
    if flag in flagType:
        desc = flagType[flag]
        for i in desc:
            flagInfo.append(flagDesc[i])
            
    return flagInfo


if __name__=='__main__':
    parser = OptionParser(description=__doc__)
    parser.add_option("-f",dest="f",help="flag",metavar="f")
        
    (options, args)=parser.parse_args()
        
    flag = options.f
    flag = int(flag)
    
    flagInfo = flagInfo(flag)
    
    for i in flagInfo:
        print i
    


        
        
        
        
    